#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <assert.h>


///////////////////////////////////////
// définition du type noeud d'un ARN
///////////////////////////////////////
struct s_nd {
	void* donnee;         // pointeur vers la donnée à stocker
	int cle;              // la clé asscoiée
	struct s_nd* fg;      // pointeur vers le fils gauche
	struct s_nd* fd;      // pointeur vers le fils droit
	bool rg;              // indique si le noeud est rouge
}; 
typedef struct s_nd nd;

//-- un arn est le pointeur vers son nd racine s'il est non vide
//-- un arn vide est le pointeur NULL
typedef nd* arn; 

//-- on crée un type enum pour garder la mémoire des orientations
enum e_orientation {
	droite, 
	gauche
};
typedef enum e_orientation orient;

char* str_orient(orient o){
	// chaine de caractère pour afficher l'orientation o
	if (o==gauche) return "gauche"; else return "droite";
} 



///////////////////////////////////////
// petites fonctions sur les arn
///////////////////////////////////////
arn cree_feuille (int c, void* d) {
	//retourne un pointeur vers un noeud de clé c, de donnée d
	//sans père ni enfant, et qui par défaut est noir
	//attention : réserve l'espace pour un noeud avec malloc
	printf("Il faut coder cree_feuille \n");
	assert(false);
}



void supprime_arn (arn a){
	//libère l'espace mémoire occupé par l'arn a 
	if (a->fg != NULL)
		supprime_arn(a->fg);
	if (a->fd != NULL)
		supprime_arn(a->fd);
	free (a);
}


bool est_vide (arn a){
	printf("Il faut coder est_vide \n");
	assert(false);
}

bool est_feuille (arn a){
	printf("Il faut coder est_feuille \n");
	assert(false);
}


///////////////////////////////////////
// tester la validité (donné ici car idem ABR, mais à savoir faire)
///////////////////////////////////////
int cle_max (arn a){
	//hyp : a!= NULL && est_valide(a) 
	//calcule la clé maximum contenue dans a
	if (a->fd == NULL){
		return a->cle;
	}
	else 
		return cle_max(a->fd);
}

int cle_min (arn a){
	//hyp : a!= NULL && est_valide(a) 
	//calcule la clé minimum contenue dans a 
	if (a->fg == NULL){
		return a->cle;
	}
	else 
		return cle_min(a->fg);
}

bool est_valide (arn a){
	//teste si a est un arn valide, ie est ordonné (=de recherche)
	//et respecte l'alternance rouge noir
	// mais on ne teste pas que la racine de l'arbre est noire...
	if (est_vide(a) || est_feuille(a)){
		return true;
	}
	else{
		bool ok_g = true;
		if (! est_vide(a->fg)){
			ok_g = est_valide (a->fg) 
				&& (cle_max(a->fg) <= a->cle)
				&& !(a->rg && a->fg->rg);
		}
		bool ok_d = true;
		if (! est_vide(a->fd)){
			ok_d = est_valide (a->fd) 
				&& (cle_min(a->fd) >= a->cle) 
				&& !(a->rg && a->fd->rg);
		}
		return ok_g && ok_d;
	}
}




///////////////////////////////////////
// recherche  (donné ici car idem ABR, mais à savoir faire)
///////////////////////////////////////
void* cherche_cle (arn a, int c){
	//renvoie un pointeur vers la/une donnée de clé c dans a
	//si un noeud de clé c existe dans a, et NULL sinon
	if (a == NULL)
		return NULL;
	if (c == a->cle)
		return a->donnee;
	if (c < a->cle)
		return cherche_cle(a->fg, c);
	else
		return cherche_cle(a->fd, c);
}





///////////////////////////////////////
// fonctions de rotation
///////////////////////////////////////
void rotation_ptr (arn u, arn v, orient ov, arn* ptr_u){
	//hyp : (u !=NULL) && (v != NULL) 
	// && ((ov==gauche && u->fg==v)  || (ov==droite && u->fd==v))
	// && (ptr_u != NULL) && *ptr_u == u
	
	// modidifie l'arn vers lequel pointe ptr_u, init. enraciné en u,
	// de sorte qu'il devient enraciné en v, et v devient père de u
	
	//printf("+++ rotation de : clé de u:%d ; clé de v:%d \n",u->cle,v->cle);
	
	printf("Il faut coder rotation_ptr \n");
	assert(false);
	
}


void rotation_pere (arn u, arn v, orient ov, arn p, orient ou){
	//hyp : (u !=NULL) && (v != NULL) && p!=NULL
	// && ((ov==gauche && u->fg==v)  || (ov==droite && u->fd==v))
	// &&((ou==gauche && p->fg==u)  || (ou==droite && p->fd==u) 
	// description à completer
	if (ou == gauche)
		rotation_ptr (u, v, ov, &p->fg);
	else
		rotation_ptr (u, v, ov, &p->fd);
}



///////////////////////////////////////
// fonction de scission
///////////////////////////////////////

bool est_4_nd (arn a){
	//hyp: a != NULL
	//retourne si le noeud racine de a représente un 4-nd
	printf("Il faut coder est_4_nd \n");
	assert(false);
}

void repare_post_fusion (arn n, arn p, orient on, 
												 arn gp, orient op, 
												 arn* ptr_agp){
	// hyp : p!=NULL && p->rg && n->rg && gp!=NULL && ptr_agp!=NULL
	//&& ((on==droite && p->fd==n) || (on==gauche && p->fg==n)))
	//&& ((op==droite && gp->fd==p)|| (op==gauche && gp->fg==p))
	//&& *ptr_agp == gp 

	// réalise les rotations necessaires pour que le 4-noeud n,p,gp
	// soit représenté par des nd rouges bien organisés au sens des ARN

	printf("Il faut coder repare_post_fusion \n");
	assert(false);
}


void scinde_ptr (arn n, arn p, orient on, 
								 arn gp, orient op, arn* ptr_gp){
	//hyp: n!= NULL && est_4_nd(n) 
	//&& (gp==NULL || (!est_nd(gp) && prt_gp!= NULL && *(ptr_gp)==gp)
	//&& ( p==NULL || (on==droite && p->fd==n) || (on=gauche && p->fg==n))
	//&& (gp==NULL || (op==droite && gp->fd==p)|| (op=gauche && gp->fg==p))
	 
	//scinde de le 4-noeud n 
	
	//printf("++++ début scission c=%d-----------------\\\n",n->cle);
	
	printf("Il faut coder scinde_ptr \n");
	assert(false);
	
	//-- changement de couleur des fils
	//-- (c'est là qu'on scinde le 4-nd)
	
	//-- si n n'est pas racine, changement de couleur de n
	//-- (on fusionne n avec son père pour limiter la hauteur) 

	//printf("**** fin scission c=%d-------------------/\n",n->cle);
}



///////////////////////////////////////
// insertion
///////////////////////////////////////
void insere (arn* ptr_a, int c, void* d){
	//ajoute dans a un noeud de clé c et de donnée d
	printf("Il faut coder insere \n");
	assert(false);
}

